//ip330Scan.cc

/********************COPYRIGHT NOTIFICATION**********************************
This software was developed under a United States Government license
described on the COPYRIGHT_UniversityOfChicago file included as part
of this distribution.
****************************************************************************/
/*
    Original Author: Jim Kowalkowski
    Date: 2/15/95
    Current Authors: Joe Sullivan and Marty Kraimer
    Converted to MPF: 12/9/98

*/

#include <vxWorks.h>
#include <iv.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
 
#include <tickLib.h>
#include <intLib.h>

#include "Reboot.h"
#include "Ip330Scan.h"
#include "WatchDog.h"
#include "IndustryPackModule.h"

int ip330ScanDebug = 0;
static int secondsBetweenCalibrate = 60;

#define TIME_PRESCALE 64	// Timer Prescale Register	64 - 255 valid
#define TIME_CONVERT  125	// Conversion Timer Register 1 - 65535 valid

/* Time between conversions in us = (timePrescale * conversionTime) / 8 */

#define NUM_AVE_VAL 200	// Number of passes through ADCs
                        // to compute value reported to EPICS.
#define SCAN_DISABLE	0xC8FF  // control register AND mask -
                                // SCAN and INTERRUPT disable
#define ACROMAG_ID 0xa3
#define ACRO_IP330 0x11

static const int nRanges=4;
static const int nGains=4;
static const char *rangeName[nRanges] = {"-5to5","-10to10","0to5","0to10"};
static const double pgaGain[nGains] = {1.0,2.0,4.0,8.0};

class ip330ADCregs
{
public:
    unsigned short control;
    unsigned char timePrescale;
    unsigned char intVector;
    unsigned short conversionTime;
    unsigned char endChanVal;
    unsigned char startChanVal;
    unsigned short newData[2];
    unsigned short missedData[2];
    unsigned short startConvert;
    unsigned char pad[0x0E];
    unsigned char gain[32];
    unsigned short mailBox[32];
};

class ip330ADCchans
{
public:
    double volt_callo;
    double volt_calhi;
    unsigned char ctl_callo;
    unsigned char ctl_calhi;
    double ideal_span;
    double ideal_zero;
    double adj_slope;
    double adj_offset;
    long chanVal;
    int gain;
    long averageStore;
};

class callibrationSetting{
public:
        double volt_callo;
	double volt_calhi;
	unsigned char ctl_callo;
	unsigned char ctl_calhi;
	double ideal_span;
	double ideal_zero;
};

static callibrationSetting callibrationSettings[nRanges][nGains] = {
    {   {0.0000, 4.9000,  0x38,  0x18, 10.0,  -5.0},
        {0.0000, 2.4500,  0x38,  0x20, 10.0,  -5.0},
        {0.0000, 1.2250,  0x38,  0x28, 10.0,  -5.0},
        {0.0000, 0.6125,  0x38,  0x30, 10.0,  -5.0} },
    {   {0.0000, 4.9000,  0x38,  0x18, 20.0, -10.0},
        {0.0000, 4.9000,  0x38,  0x18, 20.0, -10.0},
        {0.0000, 2.4500,  0x38,  0x20, 20.0, -10.0},
        {0.0000, 1.2250,  0x38,  0x28, 20.0, -10.0} },
    {   {0.6125, 4.9000,  0x30,  0x18,  5.0,   0.0},
        {0.6125, 2.4500,  0x30,  0x20,  5.0,   0.0},
        {0.6125, 1.2250,  0x30,  0x28,  5.0,   0.0},
        {0.0000, 0.6125,  0x38,  0x30,  5.0,   0.0} },
    {   {0.6125, 4.9000,  0x30,  0x18, 10.0,   0.0},
        {0.6125, 4.9000,  0x30,  0x18, 10.0,   0.0},
        {0.6125, 2.4500,  0x30,  0x20, 10.0,   0.0},
        {0.6125, 1.2250,  0x30,  0x28, 10.0,   0.0} }
};

Ip330Scan * Ip330Scan::init(
    const char *moduleName, const char *carrierName, const char *siteName,
    const char *typeString,const char *rangeString, int intVec)
{
    IndustryPackModule *pIPM = IndustryPackModule::createIndustryPackModule(
        moduleName,carrierName,siteName);
    if(!pIPM) return(0);
    unsigned char manufacturer = pIPM->getManufacturer();
    unsigned char model = pIPM->getModel();
    if(manufacturer!=ACROMAG_ID) {
        printf("initIp330Scan manufacturer 0x%x not ACROMAG_ID\n",
            manufacturer);
        return(0);
    }
    if(model!=ACRO_IP330) {
       printf("initIp330Scan model 0x%x not a ACRO_IP330\n",model);
       return(0);
    }
    signalType type;
    if(::strcmp(typeString,"D")==0) {
        type = differential;
    } else if(::strcmp(typeString,"S")==0) {
        type = singleEnded;
    } else {
        printf("Illegal type. Must be \"D\" or \"S\"\n");
        return(0);
    }
    int range;
    for(range=0; range<nRanges; range++) {
        if(::strcmp(rangeString,rangeName[range])==0) break;
    }
    if(range>=nRanges) {
        printf("Illegal range\n");
        return(0);
    }
    Ip330Scan *pIp330Scan = new Ip330Scan(pIPM,type,range,intVec);
    return(pIp330Scan);
}

Ip330Scan:: Ip330Scan(
    IndustryPackModule *pIndustryPackModule,
    signalType type, int range, int intVec)
: pIPM(pIndustryPackModule), type(type), range(range),
  accumulated(0), rebooting(0)
{
    chanData = (ip330ADCchans *)calloc(32,sizeof(ip330ADCchans));
    if(!chanData) {
        printf("Ip330Scan calloc failed\n");
        return;
    }
    regs = (ip330ADCregs *) pIPM->getMemBaseIO();
    wdId = new WatchDog;
    printf("Ip330Scan regs %p\n",regs);
    regs->startConvert = 0x0000;
    pIPM->intConfig(0); pIPM->intConfig(1);
    if(intConnect(INUM_TO_IVEC(intVec),(VOIDFUNCPTR)intFunc,(int)this)==ERROR){
        printf("Ip330Scan intConnect Failure\n");
    }
    Reboot::rebootHookAdd(rebootCallback,(void *)this);
    pIPM->intDisable(0);
    pIPM->intClear(0);
    /* Program chip registers */
    regs->control = 0x0000;
    regs->control |= 0x0002; /* Output Data Format = Straight Binary */
    /* External Trigger = Input */
    regs->control |= 0x0300; /* Scan Mode = Burst Continuous */
    regs->control |= 0x0800; /* Timer Enable = Enable */
    regs->control |= 0x2000; /* = Interrupt After All Selected */
    regs->timePrescale = TIME_PRESCALE;
    regs->conversionTime = TIME_CONVERT;
    /* Time between conversions in us = (timePrescale * conversionTime) / 8 */
    regs->startChanVal = 0x00;
    if(type==differential) {
        lastChan = 15;
        regs->control |= 0x0000;
    } else {
        lastChan = 31;
        regs->control |= 0x0008;
    }
    /* Convert all channels*/
    regs->endChanVal = lastChan;
    for (int i = 0; i <= lastChan; i++) {
        // default to gain of 0
        setGain(range,0,i);
    }
    regs->intVector = intVec;
    wdId->start(secondsBetweenCalibrate,autoCalibrate,(void *)this);
}

int Ip330Scan::getValue(int channel)
{
    return( (int) (chanData[channel].adj_slope *
        (((double)chanData[channel].chanVal + chanData[channel].adj_offset))));
}


int Ip330Scan::setGain(int gain,int channel)
{
    if(gain<0 || gain>nGains || channel<0 || channel>lastChan) return(-1);
    if(gain==chanData[channel].gain) return(0);
    return(setGain(range,gain,channel));
}

void Ip330Scan:: intFunc(void *v)
{
    Ip330Scan        *t = (Ip330Scan *) v;
    int             i;
    int             lastChannel;
    unsigned short  ad;

    lastChannel = (int) (t->regs->endChanVal);

    t->pIPM->intDisable(0);
    t->pIPM->intClear(0);
    t->accumulated++;
    for (i = 0; i <= lastChannel; i++) {
	ad = (t->regs->mailBox[i]);
	(t->chanData[i].averageStore) += ad;
	if (t->accumulated == NUM_AVE_VAL) {
	    t->chanData[i].chanVal = 
                t->chanData[i].averageStore / t->accumulated;
	    t->chanData[i].averageStore = 0;
	}
    }
    if(t->accumulated==NUM_AVE_VAL) t->accumulated = 0;
    if(!t->rebooting) t->pIPM->intEnable(0);
}

int Ip330Scan:: setGain(int range, int gain, int channel)
{
    unsigned short  saveControl;
    if(gain<0 || gain>=nGains) {
        printf("Ip330Scan setGain illegal gain value %d\n",gain);
        return(-1);
    }
    saveControl = regs->control;
    regs->control &= SCAN_DISABLE;
    chanData[channel].gain = gain;
    chanData[channel].volt_callo = callibrationSettings[range][gain].volt_callo;
    chanData[channel].volt_calhi = callibrationSettings[range][gain].volt_calhi;
    chanData[channel].ctl_callo = callibrationSettings[range][gain].ctl_callo;
    chanData[channel].ctl_calhi = callibrationSettings[range][gain].ctl_calhi;
    chanData[channel].ideal_span = callibrationSettings[range][gain].ideal_span;
    chanData[channel].ideal_zero = callibrationSettings[range][gain].ideal_zero;
    regs->gain[channel] = gain;
    regs->control = saveControl;
    calibrate(channel);
    return(0);
}

#define MAX_TIMES 1000000
void Ip330Scan:: waitNewData()
{
    int ntimes=0;
    while(ntimes++ < MAX_TIMES) {
        if(regs->newData[1]==0xffff) return;
    }
    printf("Ip330Scan::waitNewData time out\n");
}

void Ip330Scan::autoCalibrate(void * parm)
{
    Ip330Scan *pIp330Scan = (Ip330Scan *)parm;
    for(int i=0; i<= pIp330Scan->lastChan; i++) pIp330Scan->calibrate(i);
    pIp330Scan->wdId->start(secondsBetweenCalibrate,autoCalibrate,parm);
}
    

//See Acrtomag User's Manual for details about callibration
int Ip330Scan:: calibrate(int channel)
{
    unsigned short saveControl = regs->control;
    regs->control &= SCAN_DISABLE;
    /* Disable scan mode and interrupts */
    pIPM->intDisable(0);
    pIPM->intClear(0);

    // determine count_callo
    unsigned char saveStartChanVal = regs->startChanVal;
    unsigned char saveEndChanVal = regs->endChanVal;
    regs->endChanVal = 0x1F;
    regs->startChanVal = 0x00;
    for (int i = 0; i < 32; i++) regs->gain[i] = chanData[channel].gain;
    regs->control = 0x0402 | (0x0038 & (chanData[channel].ctl_callo));
    regs->startConvert = 0x0001;
    waitNewData();
    long sum = 0;
    for (int i = 0; i < 32; i++) sum = sum + regs->mailBox[i];
    double count_callo = sum/32;

    // determine count_calhi
    regs->control = 0x0402 | (0x0038 & (chanData[channel].ctl_calhi));
    regs->startConvert = 0x0001;
    waitNewData();
    sum = 0;
    for (int i = 0; i < 32; i++) sum = sum + regs->mailBox[i];
    double count_calhi = sum/32;

    double m = pgaGain[chanData[channel].gain] *
        ((chanData[channel].volt_calhi - chanData[channel].volt_callo) /
         (count_calhi - count_callo));
    double cal1 = (65536.0 * m) / chanData[channel].ideal_span;
    double cal2 =
          ((chanData[channel].volt_callo * pgaGain[chanData[channel].gain])
            - chanData[channel].ideal_zero)
          / m - count_callo;
    chanData[channel].adj_slope = cal1;
    chanData[channel].adj_offset = cal2;

    //restore control and gain values
    regs->control &= SCAN_DISABLE;
    //Disable scan and interrupts
    regs->control = saveControl;
    //Restore pre - calibrate control register state
    regs->startChanVal = saveStartChanVal;
    regs->endChanVal = saveEndChanVal;
    for (int i = 0; i < 32; i++) regs->gain[i] = chanData[i].gain;

    if(!rebooting) pIPM->intEnable(0);
    regs->startConvert = 0x0001;
    return (0);

}

void Ip330Scan:: rebootCallback(void *v)
{
    Ip330Scan *pIp330Scan = (Ip330Scan *)v;
    pIp330Scan->rebooting = true;
    pIp330Scan->pIPM->intDisable(0); pIp330Scan->pIPM->intDisable(1);
}
