// serialServer.cc

// Author: Marty Kraimer 
// Date: 7/98


#include <vxWorks.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <taskLib.h>
#include <semLib.h>
 
#include "Message.h"
#include "Int32Message.h"
#include "Char8ArrayMessage.h"
#include "SerialConfigMessage.h"
#include "serialServer.h"
#include "WatchDog.h"
#include "SerialPort.h"
 
static const int readRetrys = 3;
 
enum readState {readIdle, readActive, gotFirstEom};
enum readStatus { readSuccess=0, readTimeout };

class Serial
{
public:
    Serial(const char *name,const char *portName,
	int bufsize, int queueSize, const unsigned char *eomstr);
    static void serialServer(Serial *);
    void processConfig(SerialConfigMessage *);
    void processChar8Array(Char8ArrayMessage *);
    bool getStartOk() const {return(startOk);}
private:
    const char *serverName;
    static byteHandlerRC byteHandler(void*,unsigned char);
    MessageServer *pMessageServer;
    SerialPort* pSerialPort;
    SEM_ID done;
    bool startOk;
    readState state;
    readStatus status;
    unsigned char eom[2];
    int lenEom;
    int nextRead;
    int readBufSize;
    unsigned char *readBuf;
};

static char taskname[] = "serial";
int initSerialServer(char *serverName,char *portName,int bufSize,
    int queueSize, const unsigned char *eomstr)
{
    if(bufSize<=0) {
	printf("%s bufSize must be >0\n",serverName);
	return(0);
    }
    Serial *pSerial =
	new Serial(serverName,portName,bufSize,queueSize, eomstr);
    if(!pSerial->getStartOk()) return(0);
    int taskId = taskSpawn(taskname,100,VX_FP_TASK,2000,
	(FUNCPTR)Serial::serialServer,(int)pSerial,0,0,0,0,0,0,0,0,0);
    if(taskId==ERROR) 
	printf("%s serialServer taskSpawn Failure\n",serverName);
    return(0);
}

Serial:: Serial(const char *name,const char *portName,
    int bufsize, int queueSize, const unsigned char *eomstr)
:   serverName(name), startOk(false), state(readIdle), status(readSuccess),
    nextRead(0), readBufSize(bufsize), readBuf(0)
  
{
    if(eomstr==0) {
	eom[0] = 0; lenEom = 1;
    } else switch(strlen((const char  *)eomstr)) {
    case 0: eom[0] = 0; lenEom = 1; break;
    case 1: eom[0] = eomstr[0]; lenEom = 1; break;
    case 2: eom[0] = eomstr[0]; eom[1] = eomstr[1]; lenEom = 2; break;
    default:
	printf("%s serialServer illegal length eomstr. eom set to 0\n",
	    serverName);
	eom[0] = 0; lenEom = 1; break;
    }
    readBuf = (unsigned char *)(new char[(size_t)readBufSize]);
    pSerialPort = SerialPort::bind(portName,this,byteHandler);
    if(pSerialPort==0) {
	printf("%s serialServer: could not bind to SerialPort\n",serverName);
        return;
    }
    done = semBCreate(SEM_Q_FIFO,SEM_EMPTY);
    if(done==NULL) {
	 printf("%s serialServer: semBCreate failed\n",serverName);
	 return;
    }
    pMessageServer = new MessageServer(name,queueSize);
    startOk = true;
}

byteHandlerRC Serial::byteHandler(void* v, unsigned char data)
{
    Serial *pSerial = (Serial *)v;
    switch(pSerial->state) {
    case readIdle: return(byteHandlerError);
    case readActive:
        pSerial->readBuf[pSerial->nextRead++] = data;
        if(data==pSerial->eom[0]) {
	    if(pSerial->lenEom==1) {
		pSerial->state=readIdle;
		return(byteHandlerEndRead);
	    }
	    pSerial->state=gotFirstEom;
	}
	break;
    case gotFirstEom:
        pSerial->readBuf[pSerial->nextRead++] = data;
	if(data==pSerial->eom[1]) {
	    pSerial->state=readIdle;
	    return(byteHandlerEndRead);
	}
	pSerial->state=readActive;
	break;
    }
    if((pSerial->nextRead)>=pSerial->readBufSize) {
	pSerial->state=readIdle;
	return(byteHandlerError);
    }
    return(byteHandlerOK);
}

void Serial::serialServer(Serial *pSerial)
{
    while(true) {
        pSerial->pMessageServer->waitForMessage();
	Message *message;
        while((message = pSerial->pMessageServer->receive())) {
            switch(message->getType()) {
	    case messageTypeSerialConfig:
	        pSerial->processConfig((SerialConfigMessage *)message);
	        break;
	    case messageTypeChar8Array:
	        pSerial->processChar8Array((Char8ArrayMessage *)message);
	        break;
            default:
                printf("%s serialServer got illegal message type %d\n",
                    pSerial->serverName, message->getType());
                break;
	    }
	    delete message;
	}
    }
    return;
}
void Serial::processConfig(SerialConfigMessage *pconfig)
{
    Int32Message *preply = (Int32Message *)pMessageServer->
	allocReplyMessage(pconfig,messageTypeInt32);
    bool result = pSerialPort->config(
	    (int)pconfig->baud,(int)pconfig->stopBits,
	    (int)pconfig->bitsPerChar,(char)pconfig->parity,
            (char)pconfig->flowControl);
    preply->status = (result ? 0 : 1);
    pMessageServer->reply(preply);
}

void Serial::processChar8Array(Char8ArrayMessage *preceive)
{
    status = readSuccess;
    cmdType cmd = (cmdType)preceive->cmd;
    if(cmd==cmdSetEom) {
        Int32Message *preply = (Int32Message *)pMessageServer->
	    allocReplyMessage(preceive,messageTypeInt32);
	int len = preceive->getSize();
	if(len<1 || len>2) {
	    printf("%s Serial::processChar8Array illegal length eom\n",
		serverName);
	    preply->status = 1;
	} else {
	    lenEom = len;
	    unsigned char *pvalue = (unsigned char *)preceive->value;
	    eom[0] = pvalue[0];
	    if(len==2) eom[1] = pvalue[1];
	    preply->status = 0;
	}
        pMessageServer->reply(preply);
	return;
    }
    serialPortStatus status = serialPortFailure;
    for(int nextTry=0; nextTry<readRetrys; nextTry++) {
	switch(cmd) {
	case cmdWrite:
	    nextRead = 0; state = readIdle;
	    status = pSerialPort->write((unsigned char *)preceive->value,
		preceive->getSize(), preceive->timeout);
	    break;
	case cmdRead:
	    nextRead = 0; state = readActive;
	    status = pSerialPort->read(preceive->timeout);
	    break;
	case cmdWriteRead:
	    nextRead = 0; state = readActive;
	    status = pSerialPort->writeRead((unsigned char *)preceive->value,
		preceive->getSize(), preceive->timeout);
	    break;
	default:
	    printf("%s Got illegal message\n",serverName);
	    break;
	}
	if(status==serialPortSuccess) break;
    }
    Message *preply;
    if(status!=serialPortSuccess) {
	Int32Message *pm = (Int32Message *)pMessageServer->
	    allocReplyMessage(preceive,messageTypeInt32);
	pm->status = status;
	preply = pm;
    } else {
        Char8ArrayMessage *pm = (Char8ArrayMessage *)pMessageServer->
            allocReplyMessage(preceive,messageTypeChar8Array);
	int len = nextRead;
	pm->allocValue(len+1);
	::memcpy((void *)pm->value,(void *)readBuf,len);
	pm->value[len] = 0; //make sure string is null terminated
	pm->setSize(len);
	preply = pm;
    }
    if(preply) pMessageServer->reply(preply);
    return;
}
